from typing import List

import h5py
from funlbm.particle import Particle, ParticleConfig
from .base import create_particle
from funlbm.util import logger


class ParticleSwarm:
    def __init__(self, configs=[], device="cpu"):
        self.particles: List[Particle] = []
        for config in configs:
            self.particles.append(
                create_particle(ParticleConfig(config_json=config), device=device)
            )

    def init(self, *args, **kwargs):
        for particle in self.particles:
            particle.init(*args, **kwargs)
        self.save_particle()

    def update(self, *args, **kwargs):
        for particle in self.particles:
            particle.update(*args, **kwargs)

    def save_particle(self):
        with h5py.File("./funlbm-particle-swarm-lagrange.h5", "w") as fw:
            for i, particle in enumerate(self.particles):
                fw.create_dataset(f"particle{i}", particle._lagrange.cpu().numpy())
        logger.success("save particle lagrange success.")



import os

import click

from funlbm.lbm import LBMD3, Config

document_url = "https://darkchat.yuque.com/org-wiki-darkchat-gfaase/ul41go"


@click.group()
def funlbm():
    pass


@funlbm.command()
@click.option("--config", default="./config.json", help=f"参数配置，{document_url}")
def run(config: str = "./config.json"):
    if not os.path.exists(config):
        info = f"""配置文件不存在，访问{document_url}去配置参数吧"""
        print(info)
        raise FileExistsError(config)
    _config = Config().from_file(config)
    lbm = LBMD3(config=_config)
    lbm.run()



import math
from typing import List

import h5py
import numpy as np
import torch
from funutil import deep_get, run_timer
from scipy.optimize import fsolve

from funlbm.base import Worker
from funlbm.config.base import BaseConfig
from funlbm.particle.coord import CoordConfig, Coordinate
from funlbm.util import logger, tensor_format


class ParticleConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.coord_config: CoordConfig = CoordConfig()
        self.type = "ellipsoid"

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.coord_config.from_json(deep_get(config_json, "coord"))
        self.type = deep_get(config_json, "type") or self.type


def cul_point(xl, yl, zl, xr, yr, zr, cul_value, dx=0.5, device="cuda"):
    # 直接使用torch计算并在GPU上运行
    x = torch.linspace(xl, xr, int((xr - xl) / dx), device=device)
    y = torch.linspace(yl, yr, int((yr - yl) / dx), device=device)
    z = torch.linspace(zl, zr, int((zr - zl) / dx), device=device)

    X, Y, Z = torch.meshgrid(x, y, z, indexing="ij")
    value = cul_value(X, Y, Z)

    # 确保使用浮点类型进行计算
    inner = (value < 0).to(dtype=torch.float32)
    outer = (value > 0).to(dtype=torch.float32) * 2
    S = inner + outer

    T = S[:-1, :-1, :-1]
    T = T + S[1:, :-1, :-1] + S[:-1, 1:, :-1] + S[:-1, :-1, 1:]
    T = T + S[:-1, 1:, 1:] + S[1:, :-1, 1:] + S[1:, 1:, :-1]
    T = T + S[1:, 1:, 1:]

    # 计算平均值
    xm = T.mean(dim=2).mean(dim=1)  # 使用dim替代axis
    ym = T.mean(dim=2).mean(dim=0)
    zm = T.mean(dim=0).mean(dim=0)

    result = []
    for i in range(T.shape[0]):
        if xm[i] == 16:
            continue
        for j in range(T.shape[1]):
            if ym[j] == 16:
                continue
            for k in range(T.shape[2]):
                if zm[k] == 16:
                    continue
                if T[i, j, k] == 16:
                    continue

                result.append([X[i][j][k].item(), Y[i][j][k].item(), Z[i][j][k].item()])

    logger.info(f"lagrange size:{len(result)}")
    return np.array(result, dtype=np.float32)


def find_intersection(point1, point2, cul_value):
    """
    找到线段与椭球表面的交点。
    point1: 线段起点 (在椭球内部)。
    point2: 线段终点 (在椭球外部)。
    """

    def ellipsoid_equation(t):
        x = point1[0] + t * (point2[0] - point1[0])
        y = point1[1] + t * (point2[1] - point1[1])
        z = point1[2] + t * (point2[2] - point1[2])
        return cul_value(x, y, z)

    # 使用数值方法求解交点
    t = fsolve(ellipsoid_equation, 0.5)[0]
    x = point1[0] + t * (point2[0] - point1[0])
    y = point1[1] + t * (point2[1] - point1[1])
    z = point1[2] + t * (point2[2] - point1[2])
    return np.array([x, y, z])


def generate_uniform_points_on_ellipsoid(
    xl, yl, zl, xr, yr, zr, cul_value, dx=0.5, *args, **kwargs
):
    """
    在椭球表面上生成均匀分布的点。

    参数:
        a (float): 椭球的x轴半轴长度。
        b (float): 椭球的y轴半轴长度。
        c (float): 椭球的z轴半轴长度。
        h (float): 空间细分的步长（方块的边长）。

    返回:
        np.ndarray: 椭球表面上的均匀分布点，形状为 (N, 3)。
    """
    # 定义椭球的包围盒范围
    x_range = np.arange(xl, xr, dx)
    y_range = np.arange(yl, yr, dx)
    z_range = np.arange(zl, zr, dx)

    # 存储椭球表面上的点
    surface_points = []

    # 遍历所有方块
    for i in range(len(x_range) - 1):
        for j in range(len(y_range) - 1):
            for k in range(len(z_range) - 1):
                # 当前方块的 8 个顶点
                x0, x1 = x_range[i], x_range[i + 1]
                y0, y1 = y_range[j], y_range[j + 1]
                z0, z1 = z_range[k], z_range[k + 1]
                vertices = [
                    (x0, y0, z0),
                    (x0, y0, z1),
                    (x0, y1, z0),
                    (x0, y1, z1),
                    (x1, y0, z0),
                    (x1, y0, z1),
                    (x1, y1, z0),
                    (x1, y1, z1),
                ]

                # 判断顶点是否在椭球内部或外部
                inside = [cul_value(x, y, z) < 0 for x, y, z in vertices]
                if any(inside) and not all(inside):
                    # 当前方块与椭球表面相交
                    # 找到所有与椭球表面相交的边
                    intersection_points = []
                    for idx1 in range(len(vertices)):
                        for idx2 in range(idx1 + 1, len(vertices)):
                            if inside[idx1] != inside[idx2]:
                                # 找到交点
                                point1 = np.array(vertices[idx1])
                                point2 = np.array(vertices[idx2])
                                intersection = find_intersection(
                                    point1, point2, cul_value
                                )
                                intersection_points.append(intersection)

                    # 计算切面的重心
                    if len(intersection_points) >= 3:
                        centroid = np.mean(intersection_points, axis=0)
                        surface_points.append(centroid)

    return np.array(surface_points)


class Particle(Worker):
    """
    粒子基类

    属性:
        config (ParticleConfig): 粒子配置
        coord (Coordinate): 坐标系对象
        mass (float): 粒子质量
        area (float): 粒子表面积
        I (Tensor): 惯性矩
        rou (float): 粒子密度
        angle (Tensor): 粒子方向
        cx (Tensor): 质心坐标 [x,y,z]
        cr (Tensor): 质心半径 [a,b,b]
        cu (Tensor): 质心速度 [vx,vy,vz]
        cw (Tensor): 质心角速度 [wx,wy,wz]
        cF (Tensor): 质心合外力
        cT (Tensor): 质心合外力矩
        lx (Tensor): 拉格朗日点坐标 [m,i,3]
        lF (Tensor): 拉格朗日点受力 [m,i,3]
        lm (Tensor): 拉格朗日点质量
        lu (Tensor): 拉格朗日点速度 [m,i,3]
        lrou (Tensor): 拉格朗日点密度
    """

    def __init__(self, config: ParticleConfig = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.config: ParticleConfig = config or ParticleConfig()
        self.coord: Coordinate = Coordinate(
            config=self.config.coord_config, *args, **kwargs
        )

        # 颗粒质量[1]
        self.mass = None
        # 颗粒面积[1]
        self.area = None
        # 惯性矩
        self.I = None
        # 颗粒密度[1]
        self.rou = None
        # 颗粒方向[i,j,k]
        self.angle = None

        # 质心坐标[i,j,k]
        self.cx = torch.tensor(
            self.config.coord_config.center, device=self.device, dtype=torch.float32
        )
        # 质心半径[a,b,b]
        self.cr = 5 * torch.ones(5, device=self.device, dtype=torch.float32)
        # 质心速度[i,j,k]
        self.cu = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心角速度[i,j,k]
        self.cw = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心合外力
        self.cF = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心合外力
        self.cT = torch.zeros(3, device=self.device, dtype=torch.float32)

        self._lagrange: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点的坐标[m,i,3]
        self.lx: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点上的力[m,i,3]
        self.lF: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点的质量
        self.lm: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点速度[m,i,3]
        self.lu: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点速度[m,i,3]
        self.lrou: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )

    def _init(self, dx=1, *args, **kwargs):
        raise NotImplementedError("还没实现")

    def init(self, *args, **kwargs):
        self.rou = float(self.config.get("rou", 1.0))
        self._init(*args, **kwargs)
        shape = self._lagrange.shape
        self.lx = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lF = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lu = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lm = torch.full(
            (shape[0], 1), self.area / shape[0], device=self.device, dtype=torch.float32
        )
        self.lrou = torch.empty((shape[0], 1), device=self.device, dtype=torch.float32)

    @run_timer
    def update_from_lar(self, dt: float, gl: float = 9.8, rouf: float = 1.0) -> None:
        """
        从拉格朗日坐标更新粒子状态。

        参数:
            dt: 时间步长
            gl: 重力加速度
            rouf: 流体密度

        异常:
            ValueError: 当输入参数无效时抛出
        """
        if dt <= 0:
            raise ValueError("Time step must be positive")
        if rouf <= 0:
            raise ValueError("Fluid density must be positive")

        tmp = (
            (1 - self.rou / rouf)
            * self.mass
            * torch.tensor([gl, 0, 0], device=self.device)
        )
        self.cF = torch.sum(-self.lF * self.lm, dim=0) + tmp

        self.cu = self.cu + self.cF / self.mass * dt
        self.cx = self.cx + self.cu * dt

        self.cT = -torch.sum(
            torch.cross(self.lx - self.cx, self.lF, dim=-1) * self.lm, dim=0
        )
        self.cw = self.cw + 0.1 * self.cT * dt / self.I

    @run_timer
    def update(self, dt):
        self.coord.update(center=self.cx, w=self.cw)
        self.lx = self.coord.cul_point(self._lagrange)

    def from_json(self):
        pass

    def to_str(self, step=0, *args, **kwargs):
        res = f"m={self.mass:.2f}"
        res += "\tcu=" + ",".join([f"{i:.6f}" for i in self.cu])
        res += "\tcx=" + ",".join([f"{i:.6f}" for i in self.cx])
        res += "\tcf=" + ",".join([f"{i:.6f}" for i in self.cF])
        res += "\tlF=" + ",".join(
            [f"{i:.6f}" for i in [self.lF.min(), self.lF.mean(), self.lF.max()]]
        )
        res += "\tcw=" + ",".join([f"{i:.6f}" for i in self.cw])
        res += "\tr=" + self.coord.to_str()
        return res

    def to_json(self, step=0, *args, **kwargs):
        return {
            "m": float(self.mass.cpu().numpy()),
            "cu": tensor_format(self.cu),
            "cx": tensor_format(self.cx),
            "cf": tensor_format(self.cF),
            "lF": tensor_format(
                [
                    self.lF.min(),
                    self.lF.mean(),
                    self.lF.max(),
                ]
            ),
            "cw": tensor_format(self.cw),
            "coord": self.coord.to_json(),
        }


class Ellipsoid(Particle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ra, self.rb, self.rc = (
            self.config.get("a") or 10,
            self.config.get("b") or 10,
            self.config.get("c") or 10,
        )

    def _init(self, dx=1, *args, **kwargs):
        xl, yl, zl = -self.ra - 2 * dx, -self.rb - 2 * dx, -self.rc - 2 * dx
        xr, yr, zr = self.ra + 2 * dx, self.rb + 2 * dx, self.rc + 2 * dx
        self._lagrange = torch.tensor(
            generate_uniform_points_on_ellipsoid(
                xl,
                yl,
                zl,
                xr,
                yr,
                zr,
                cul_value=lambda X, Y, Z: X**2 / self.ra**2
                + Y**2 / self.rb**2
                + Z**2 / self.rc**2
                - 1,
                dx=dx,
            ),
            device=self.device,
            dtype=torch.float32,
        )
        self.mass = torch.tensor(
            4.0 / 3.0 * math.pi * self.ra * self.rb * self.rc,
            device=self.device,
            dtype=torch.float32,
        )
        self.area = torch.tensor(
            4.0 / 3.0 * math.pi * math.pow(self.ra * self.rb * self.rc, 2.0 / 3.0),
            device=self.device,
            dtype=torch.float32,
        )
        self.I = torch.tensor(
            np.array([self.rb * self.rc, self.ra * self.rc, self.ra * self.rb])
            * self.mass.to("cpu").numpy()
            / 5.0,
            device=self.device,
            dtype=torch.float32,
        )


class Sphere(Particle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.r = self.config.get("r") or 10

    def _init(self, dx=1, *args, **kwargs):
        xl, yl, zl = -self.r - 2 * dx, -self.r - 2 * dx, -self.r - 2 * dx
        xr, yr, zr = self.r + 2 * dx, self.r + 2 * dx, self.r + 2 * dx
        self._lagrange = torch.tensor(
            generate_uniform_points_on_ellipsoid(
                xl,
                yl,
                zl,
                xr,
                yr,
                zr,
                cul_value=lambda X, Y, Z: X**2 / self.r**2
                + Y**2 / self.r**2
                + Z**2 / self.r**2
                - 1,
                dx=dx,
                device=self.device,
            ),
            device=self.device,
            dtype=torch.float32,
        )
        self.mass = torch.tensor(
            4.0 / 3.0 * math.pi * self.r * self.r * self.r,
            device=self.device,
            dtype=torch.float32,
        )
        self.area = torch.tensor(
            4.0 * math.pi * self.r * self.r,
            device=self.device,
            dtype=torch.float32,
        )

        self.I = torch.tensor(
            np.array([self.r * self.r, self.r * self.r, self.r * self.r])
            * self.mass.to("cpu").numpy()
            / 5.0,
            device=self.device,
            dtype=torch.float32,
        )


def create_particle(config: ParticleConfig, device="cpu") -> Particle:
    if config.type == "ellipsoid":
        return Ellipsoid(config=config, device=device)
    else:
        return Sphere(config=config, device=device)


def example():
    ellipsoid = Ellipsoid(config=ParticleConfig())
    ellipsoid.init()
    ellipsoid.update()

    print(ellipsoid.lx)
    print("#####")
    print(ellipsoid.lx)



import math

import torch
from funutil.cache import cache

from funlbm.base import Worker


class Param(Worker):
    """格子玻尔兹曼模型参数基类

    提供离散速度、权重系数等基本参数的管理功能

    Args:
        e: 离散速度向量集合
        w: 权重系数集合
        vertex_reverse: 速度方向映射关系

    属性:
        e (Tensor): 离散速度向量
        w (Tensor): 权重系数
        cs (Tensor): 声速
        vertex_reverse (list): 速度方向映射表
    """

    def __init__(self, e, w, vertex_reverse, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vertex_reverse = vertex_reverse
        self.e = torch.tensor(e, device=self.device, dtype=torch.float32)
        self.w = torch.tensor(w, device=self.device, dtype=torch.float32)
        self.cs = torch.tensor(
            math.sqrt(1.0 / 3), device=self.device, dtype=torch.float32
        )

    @cache
    def eT(self) -> torch.Tensor:
        """返回离散速度向量的转置矩阵"""
        return self.e.t()

    @cache
    def vertex_index(self, axis, value):
        """获取指定方向和值的速度索引

        Args:
            axis: 坐标轴(0,1,2)
            value: 速度值

        Returns:
            list: 满足条件的速度索引列表
        """
        return [i for i, e in enumerate(self.e) if e[axis] == value]

    def index_reverse(self, index):
        """获取速度方向的反向索引

        Args:
            index: 速度索引,可以是单个索引或索引列表

        Returns:
            反向速度的索引
        """
        if isinstance(index, int):
            return self.vertex_reverse[index]
        return [self.vertex_reverse[i] for i in index]

    @property
    def e_dim(self):
        """离散速度的维度"""
        return self.e.shape[0]



from .base import Param
from .param3d import ParamD3Q13, ParamD3Q15, ParamD3Q19, ParamD3Q27, parse_3d_param

__all__ = [
    "Param",
    "ParamD3Q19",
    "ParamD3Q13",
    "parse_3d_param",
    "ParamD3Q15",
    "ParamD3Q27",
]



import torch

from funlbm.util import logger


def device_detect(device=None):
    """检测并返回可用的计算设备

    按优先级顺序检测:CUDA > MPS > CPU

    Args:
        device: 指定的设备名称,如果为None则自动检测

    Returns:
        torch.device: 返回可用的计算设备(cuda/mps/cpu)
    """
    if device is not None and device != "auto":
        return device
    try:
        if torch.cuda.is_available():
            return torch.device("cuda")
    except AttributeError:
        logger.error("No cuda available")
    try:
        if torch.mps.is_available():
            return torch.device("mps")
    except AttributeError:
        logger.error("No mps available")
    return torch.device("cpu")


class Worker:
    """基础工作类

    提供设备检测和管理功能

    Args:
        device: 计算设备名称,默认为"cpu"
    """

    def __init__(self, device="cpu", *args, **kwargs):
        self.device = device_detect(device)
        logger.info(f"init {type(self).__name__} with device={self.device}")



import numpy as np
import torch
from funtecplot.dump import PointData
from funutil import run_timer
from funvtk.hl import gridToVTK, pointsToVTK

from funlbm.flow import FlowD3
from funlbm.particle import ParticleSwarm
from funlbm.util import logger

from funlbm.lbm.base import Config, LBMBase


class LBMD3(LBMBase):
    """3D格子玻尔兹曼方法实现类

    实现了3D流场的基本功能,包括初始化、边界处理等

    Args:
        config: LBM配置对象
    """

    def __init__(self, config: Config, *args, **kwargs):
        flow = FlowD3(config=config.flow_config, device=config.device, *args, **kwargs)
        super(LBMD3, self).__init__(
            flow=flow,
            config=config,
            device=config.device,
            particle_swarm=ParticleSwarm(config.particles, device=config.device),
            *args,
            **kwargs,
        )

    def init(self):
        """初始化流场和颗粒"""
        # 初始化流程
        # 初始化边界条件
        self.flow.init()
        # 初始化颗粒
        self.flow.cul_equ()
        self.particle_swarm.init()

    def _calculate_region_bounds(self, particle, n=2, h=1):
        """Calculate region bounds for particle interaction."""
        # 保持在GPU上计算
        rl = torch.floor(particle.lx) - (n - 1) * h
        # 限制下界
        rl = torch.clamp(rl, min=0)
        rr = rl + (2 * n - 1) * h + 1
        # 限制上界
        domain_size = torch.tensor(
            self.config.flow_config.size, device=self.device, dtype=torch.float32
        )
        rr = torch.clamp(rr, max=domain_size)
        return rl.to(dtype=torch.int32), rr.to(dtype=torch.int32)

    def _calculate_weight_function(
        self,
        r,
        n=2,
        h=1,
    ):
        """Calculate weight function for particle-fluid interaction."""

        # 限制权重函数的作用范围
        mask = torch.all(torch.abs(r) <= n * h, dim=-1, keepdim=True)
        tmp = torch.where(
            mask, (1 + torch.cos(torch.abs(r * np.pi / 2 / h))) / 4 / h, 0.0
        )
        w = torch.prod(tmp, dim=-1, keepdim=True).to(dtype=torch.float32)
        # 归一化权重
        w_sum = torch.sum(w)
        if w_sum > 0:
            w = w / w_sum
        return w

    @run_timer
    def flow_to_lagrange(self, *args, **kwargs):
        for particle in self.particle_swarm.particles:
            rl, rr = self._calculate_region_bounds(particle)
            lu = torch.zeros_like(particle.lu, device=self.device)
            lrou = torch.zeros_like(particle.lrou, device=self.device)

            for index, lar in enumerate(particle.lx):
                il, ir = rl[index], rr[index]
                index_range = (
                    slice(il[0], ir[0]),
                    slice(il[1], ir[1]),
                    slice(il[2], ir[2]),
                    slice(None),
                )

                region_x = self.flow.x[index_range]
                tmp = self._calculate_weight_function(region_x - lar)
                if torch.abs(torch.sum(tmp) - 1) > 1e-3 and self.step > 100:
                    logger.error(f"sum={torch.sum(tmp)},异常了，程序停止")
                    self.run_status = False

                lu[index, :] = torch.sum(self.flow.u[index_range] * tmp)
                lrou[index, :] = torch.sum(self.flow.rou[index_range] * tmp)
            u_theta = 0

            particle.lu[:, :] = (
                particle.cu
                + torch.cross(
                    particle.cw.unsqueeze(0), particle.lx - particle.cx, dim=-1
                )
                + u_theta
            )

            particle.lF = 2 * lrou * (particle.lu - lu) / self.config.dt

    @run_timer
    def lagrange_to_flow(self, *args, **kwargs):
        self.flow.FOL[:, :, :, :] = 0

        for particle in self.particle_swarm.particles:
            rl, rr = self._calculate_region_bounds(particle)
            for index, lar in enumerate(particle.lx):
                il, ir = rl[index], rr[index]
                index_range = (
                    slice(il[0], ir[0]),
                    slice(il[1], ir[1]),
                    slice(il[2], ir[2]),
                    slice(None),
                )
                tmp = self._calculate_weight_function(self.flow.x[index_range] - lar)
                self.flow.FOL[index_range] = (
                    tmp * particle.lF[index, :] * particle.lm[index]
                )

    @run_timer
    def particle_to_wall(self, *args, **kwargs):
        """
        Handle particle-wall collisions using a spring force model.
        """
        k0 = 300  # Spring constant
        wall_normals = torch.tensor(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0], [0, -1, 0], [0, 0, -1]],
            device=self.device,
            dtype=torch.float32,
        )

        domain_bounds = torch.concatenate(
            [
                torch.zeros(self.config.flow_config.size.shape, device=self.device),
                torch.tensor(
                    self.config.flow_config.size,
                    device=self.device,
                    dtype=torch.float32,
                ),
            ]
        )

        for particle in self.particle_swarm.particles:
            # Find extreme points of particle
            extreme_indices = torch.concatenate(
                [torch.argmin(particle.lx, dim=0), torch.argmax(particle.lx, dim=0)]
            )

            # Calculate wall distances
            distances = k0 * (
                1
                - abs(
                    particle.lx[extreme_indices][[0, 1, 2, 3, 4, 5], [0, 1, 2, 0, 1, 2]]
                    - domain_bounds
                )
                / (2 * self.config.dx)
            )
            distances = torch.clamp(distances, min=0)

            if distances.max() == 0:
                continue

            logger.info(f"Distance of particle to wall = {distances}")
            particle.lF[extreme_indices] += torch.multiply(
                wall_normals, distances.unsqueeze(1)
            )

    def save(self, step=10, *args, **kwargs):
        if step % self.config.file_config.per_steps > 0:
            return

        shape = self.flow.u.shape
        xf, yf, zf = np.meshgrid(
            range(shape[0]),
            range(shape[1]),
            range(shape[2]),
            indexing="ij",
            sparse=False,
        )
        flow_u = self.flow.u.to("cpu").numpy()
        point_data = {
            "u": (
                flow_u[:, :, :, 0],
                flow_u[:, :, :, 1],
                flow_u[:, :, :, 2],
            ),
        }

        cell_data = {
            "rho": self.flow.rou.to("cpu").numpy()[:, :, :, 0],
            "p": self.flow.p.to("cpu").numpy()[:, :, :, 0],
        }

        # t1 = self.flow.rou[10:-10, 10:-10, 10:-10, :]
        # t1 = cell_data['rho'][5:-5, 5:-5, 5:-5]
        # print("rho", step, t1.max(), t1.min(), t1.max() - t1.min())
        PointData(
            data=flow_u,
            variables=["x", "y", "z", "ux", "uy", "uz"],
            axis_dim=3,
            data_dim=3,
            title="flow_u",
        ).dump(
            filepath=f"{self.config.file_config.tecplot_path}/tecplot_flow_u_{str(step).zfill(10)}.dat"
        )
        PointData(
            data=cell_data["rho"],
            variables=["x", "y", "z", "rho"],
            axis_dim=3,
            data_dim=1,
            title="flow_u",
        ).dump(
            filepath=f"{self.config.file_config.tecplot_path}/tecplot_flow_rho_{str(step).zfill(10)}.dat"
        )
        PointData(
            data=cell_data["p"],
            variables=["x", "y", "z", "p"],
            axis_dim=3,
            data_dim=1,
            title="flow_u",
        ).dump(
            filepath=f"{self.config.file_config.tecplot_path}/tecplot_flow_p_{str(step).zfill(10)}.dat"
        )

        gridToVTK(
            f"{self.config.file_config.vtk_path}/flow_" + str(step).zfill(10),
            xf,
            yf,
            zf,
            pointData=point_data,
            cellData=cell_data,
        )

        for i, particle in enumerate(self.particle_swarm.particles):
            lx = particle.lx.to("cpu").numpy()
            xf, yf, zf = lx[:, 0], lx[:, 1], lx[:, 2]
            data = {
                "u": particle.lu.to("cpu").numpy()[:, 0],
            }
            fname = f"{self.config.file_config.vtk_path}/particle_{str(i).zfill(3)}_{str(step).zfill(10)}"
            pointsToVTK(fname, xf, yf, zf, data=data)

        # write_to_tecplot(cell_data["rho"], f"{self.config.file_config.vtk_path}/tecplot_{str(step).zfill(10)}.dat")


class LBMD3Q27(LBMD3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LBMD3Q19(LBMD3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LBMD3Q15(LBMD3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LBMD3Q13(LBMD3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import numpy as np
import torch
from funutil import deep_get
from torch import Tensor

from funlbm.base import Worker
from funlbm.config.base import BaseConfig, BoundaryConfig
from funlbm.parameter import Param
from funlbm.util import tensor_format


class FlowConfig(BaseConfig):
    """流体配置类

    Attributes:
        size: 计算域大小,形状为(3,)的数组
        param: 参数字典
        param_type: 参数类型,默认为"D3Q19"
        boundary: 边界配置
        gl: 重力加速度
        Re: 雷诺数
        mu: 动力粘度
    """

    def __init__(self, *args, **kwargs):
        self.size: np.ndarray = np.zeros(3)  # 计算域大小
        self.param: dict = {}  # 参数字典
        self.param_type: str = "D3Q19"  # 参数类型
        self.boundary: BoundaryConfig = None  # 边界配置

        self.gl: float = 0.0  # 重力加速度
        self.Re: float = 10  # 雷诺数
        self.mu: float = 10  # 动力粘度
        super().__init__(*args, **kwargs)

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.size = np.array(
            deep_get(config_json, "size") or [100, 100, 100], dtype=int
        )
        self.param = deep_get(config_json, "param") or self.param
        self.boundary = BoundaryConfig().from_json(
            deep_get(config_json, "boundary") or {}
        )
        self.param_type = deep_get(config_json, "param_type") or self.param_type

        self.Re = float(deep_get(self.param, "Re") or self.Re)
        self.mu = float(deep_get(self.param, "mu") or self.mu)
        self.gl = float(deep_get(self.param, "gl") or self.gl)


class Flow(Worker):
    """流场基类

    实现了流场计算的基本功能

    属性:
        param (Param): 参数对象
        config (FlowConfig): 配置对象
        x (Tensor): 坐标张量
        f (Tensor): 力密度分布函数
        feq (Tensor): 平衡态分布函数
        u (Tensor): 速度场
        p (Tensor): 压力场
        rou (Tensor): 密度场
        gama (Tensor): 剪切率
        FOL (Tensor): 力项
        tau (Tensor): 松弛时间
    """

    def __init__(self, param: Param, config: FlowConfig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.param: Param = param
        self.config: FlowConfig = config

        # 坐标
        self.x: Tensor = torch.zeros([1])
        # 力密度
        self.f: Tensor = torch.zeros([1])
        self.feq: Tensor = torch.zeros([1])
        # 速度
        self.u: Tensor = torch.zeros([1])
        # 压强
        self.p: Tensor = torch.zeros([1])
        # 密度
        self.rou: Tensor = torch.zeros([1])
        # 剪切率相关的变量
        self.gama: Tensor = torch.zeros([1])

        self.FOL: Tensor = torch.zeros([1])

        self.tau: Tensor = torch.zeros([1])

    def init(self, *args, **kwargs) -> None:
        """初始化流场"""
        raise NotImplementedError("not implemented")

    def update_u_rou(self, *args, **kwargs) -> None:
        """更新速度和密度场"""
        raise NotImplementedError("not implemented")

    def cul_equ(self, tau: Tensor = None, *args, **kwargs) -> None:
        """计算平衡态分布函数"""
        raise NotImplementedError("not implemented")

    def cul_equ2(self, *args, **kwargs) -> None:
        """计算第二种平衡态分布函数"""
        raise NotImplementedError("not implemented")

    def f_stream(self, *args, **kwargs) -> None:
        """执行流动计算"""
        raise NotImplementedError("not implemented")

    def __repr__(self) -> str:
        """返回流场对象的字符串表示"""
        return f"{self.__class__.__name__}(size={tuple(self.config.size)}, Re={self.config.Re}, mu={self.config.mu})"

    def to_json(self):
        return {
            "f": tensor_format(
                [
                    self.f.min(),
                    self.f.mean(),
                    self.f.max(),
                ]
            ),
            "u": tensor_format(
                [
                    self.u.min(),
                    self.u.mean(),
                    self.u.max(),
                ]
            ),
            "rho": tensor_format(
                [
                    self.rou.min(),
                    self.rou.mean(),
                    self.rou.max(),
                ]
            ),
        }
